@@ -17,8 +17,17 @@ import os.path as osp
 import sys
 import site
 
+from traitlets import DottedObjectName
+import ipykernel
+
+# Local imports
+from spyder_kernels.utils.misc import is_module_installed
+from spyder_kernels.utils.mpl import (
+    MPL_BACKENDS_FROM_SPYDER, INLINE_FIGURE_FORMATS)
+
 
 PY2 = sys.version[0] == '2'
+IPYKERNEL_6 = ipykernel.__version__[0] >= '6'
 
 
 def import_spydercustomize():
@@ -36,7 +45,7 @@ def import_spydercustomize():
 
     # Import our customizations
     site.addsitedir(customize_dir)
-    import spydercustomize
+    import spydercustomize  # noqa
 
     # Remove our customize path from sys.path
     try:
@@ -45,19 +54,6 @@ def import_spydercustomize():
         pass
 
 
-def is_module_installed(module_name):
-    """
-    Simpler version of spyder.utils.programs.is_module_installed
-    to improve startup time.
-    """
-    try:
-        __import__(module_name)
-        return True
-    except:
-        # Module is not installed
-        return False
-
-
 def sympy_config(mpl_backend):
     """Sympy configuration"""
     if mpl_backend is not None:
@@ -107,16 +103,38 @@ def kernel_config():
     if not PY2:
         spy_cfg.IPCompleter.use_jedi = jedi_o
 
+    # Clear terminal arguments input.
+    # This needs to be done before adding the exec_lines that come from
+    # Spyder, to avoid deleting the sys module if users want to import
+    # it through them.
+    # See spyder-ide/spyder#15788
+    clear_argv = "import sys; sys.argv = ['']; del sys"
+    spy_cfg.IPKernelApp.exec_lines = [clear_argv]
+
+    # Set our runfile in builtins here to prevent other packages shadowing it.
+    # This started to be a problem since IPykernel 6.3.0.
+    if not PY2:
+        spy_cfg.IPKernelApp.exec_lines.append(
+            "import builtins; "
+            "builtins.runfile = builtins.spyder_runfile; "
+            "del builtins.spyder_runfile; del builtins"
+        )
+
+    # Prevent other libraries to change the breakpoint builtin.
+    # This started to be a problem since IPykernel 6.3.0.
+    if sys.version_info[0:2] >= (3, 7):
+        spy_cfg.IPKernelApp.exec_lines.append(
+            "import sys; import pdb; "
+            "sys.breakpointhook = pdb.set_trace; "
+            "del sys; del pdb"
+        )
+
     # Run lines of code at startup
     run_lines_o = os.environ.get('SPY_RUN_LINES_O')
     if run_lines_o is not None:
-        spy_cfg.IPKernelApp.exec_lines = [x.strip() for x in run_lines_o.split(';')]
-    else:
-        spy_cfg.IPKernelApp.exec_lines = []
-
-    # Clean terminal arguments input
-    clear_argv = "import sys;sys.argv = [''];del sys"
-    spy_cfg.IPKernelApp.exec_lines.append(clear_argv)
+        spy_cfg.IPKernelApp.exec_lines += (
+            [x.strip() for x in run_lines_o.split(';')]
+        )
 
     # Load %autoreload magic
     spy_cfg.IPKernelApp.exec_lines.append(
@@ -136,56 +154,36 @@ def kernel_config():
     else:
         dpi_option = 'figure.dpi'
 
-    spy_cfg.InlineBackend.rc = {'figure.figsize': (6.0, 4.0),
-                                dpi_option: 72,
-                                'font.size': 10,
-                                'figure.subplot.bottom': .125,
-                                'figure.facecolor': 'white',
-                                'figure.edgecolor': 'white'}
+    # The typical default figure size is too large for inline use,
+    # so we shrink the figure size to 6x4, and tweak fonts to
+    # make that fit.
+    spy_cfg.InlineBackend.rc = {
+        'figure.figsize': (6.0, 4.0),
+        # 72 dpi matches SVG/qtconsole.
+        # This only affects PNG export, as SVG has no dpi setting.
+        dpi_option: 72,
+        # 12pt labels get cutoff on 6x4 logplots, so use 10pt.
+        'font.size': 10,
+        # 10pt still needs a little more room on the xlabel
+        'figure.subplot.bottom': .125,
+        # Play nicely with any background color.
+        'figure.facecolor': 'white',
+        'figure.edgecolor': 'white'
+    }
 
     # Pylab configuration
     mpl_backend = None
-    pylab_o = os.environ.get('SPY_PYLAB_O')
-
-    if pylab_o == 'True' and is_module_installed('matplotlib'):
-        # Set Matplotlib backend
+    if is_module_installed('matplotlib'):
+        # Set Matplotlib backend with Spyder options
+        pylab_o = os.environ.get('SPY_PYLAB_O')
         backend_o = os.environ.get('SPY_BACKEND_O')
-        if backend_o is not None:
-            if backend_o == '1':
-                if is_module_installed('PyQt5'):
-                    auto_backend = 'qt5'
-                elif is_module_installed('PyQt4'):
-                    auto_backend = 'qt4'
-                elif is_module_installed('_tkinter'):
-                    auto_backend = 'tk'
-                else:
-                    auto_backend = 'inline'
-            else:
-                auto_backend = ''
-            backends = {'0': 'inline',
-                        '1': auto_backend,
-                        '2': 'qt5',
-                        '3': 'qt4',
-                        '4': 'osx',
-                        '5': 'gtk3',
-                        '6': 'gtk',
-                        '7': 'wx',
-                        '8': 'tk'}
-            mpl_backend = backends[backend_o]
-
-            # Automatically load Pylab and Numpy, or only set Matplotlib
-            # backend
-            autoload_pylab_o = os.environ.get('SPY_AUTOLOAD_PYLAB_O') == 'True'
-            command = "get_ipython().kernel._set_mpl_backend('{0}', {1})"
-            spy_cfg.IPKernelApp.exec_lines.append(
-                command.format(mpl_backend, autoload_pylab_o))
-
+        if pylab_o == 'True' and backend_o is not None:
+            mpl_backend = MPL_BACKENDS_FROM_SPYDER[backend_o]
             # Inline backend configuration
             if mpl_backend == 'inline':
                 # Figure format
                 format_o = os.environ.get('SPY_FORMAT_O')
-                formats = {'0': 'png',
-                           '1': 'svg'}
+                formats = INLINE_FIGURE_FORMATS
                 if format_o is not None:
                     spy_cfg.InlineBackend.figure_format = formats[format_o]
 
@@ -206,6 +204,17 @@ def kernel_config():
                 bbox_inches = 'tight' if bbox_inches_o == 'True' else None
                 spy_cfg.InlineBackend.print_figure_kwargs.update(
                     {'bbox_inches': bbox_inches})
+        else:
+            # Set Matplotlib backend to inline for external kernels.
+            # Fixes issue 108
+            mpl_backend = 'inline'
+
+        # Automatically load Pylab and Numpy, or only set Matplotlib
+        # backend
+        autoload_pylab_o = os.environ.get('SPY_AUTOLOAD_PYLAB_O') == 'True'
+        command = "get_ipython().kernel._set_mpl_backend('{0}', {1})"
+        spy_cfg.IPKernelApp.exec_lines.append(
+            command.format(mpl_backend, autoload_pylab_o))
 
     # Enable Cython magic
     run_cython = os.environ.get('SPY_RUN_CYTHON') == 'True'
@@ -216,7 +225,8 @@ def kernel_config():
     use_file_o = os.environ.get('SPY_USE_FILE_O')
     run_file_o = os.environ.get('SPY_RUN_FILE_O')
     if use_file_o == 'True' and run_file_o is not None:
-        spy_cfg.IPKernelApp.file_to_run = run_file_o
+        if osp.exists(run_file_o):
+            spy_cfg.IPKernelApp.file_to_run = run_file_o
 
     # Autocall
     autocall_o = os.environ.get('SPY_AUTOCALL_O')
@@ -236,6 +246,11 @@ def kernel_config():
         lines = sympy_config(mpl_backend)
         spy_cfg.IPKernelApp.exec_lines.append(lines)
 
+    # Disable the new mechanism to capture and forward low-level output
+    # in IPykernel 6. For that we have Wurlitzer.
+    if LooseVersion(ipykernel.__version__) >= LooseVersion('6.3.0'):
+        spy_cfg.IPKernelApp.capture_fd_output = False
+
     # Merge IPython and Spyder configs. Spyder prefs will have prevalence
     # over IPython ones
     cfg._merge(spy_cfg)
@@ -255,10 +270,9 @@ def varexp(line):
         import guiqwt.pyplot as pyplot
     except:
         import matplotlib.pyplot as pyplot
-    __fig__ = pyplot.figure();
-    __items__ = getattr(pyplot, funcname[2:])(ip.user_ns[name])
+    pyplot.figure();
+    getattr(pyplot, funcname[2:])(ip.kernel._get_current_namespace()[name])
     pyplot.show()
-    del __fig__, __items__
 
 
 def main():
@@ -285,11 +299,27 @@ def main():
     while '' in sys.path:
         sys.path.remove('')
 
-    # Fire up the kernel instance.
+    # Main imports
     from ipykernel.kernelapp import IPKernelApp
     from spyder_kernels.console.kernel import SpyderKernel
 
-    kernel = IPKernelApp.instance()
+    class SpyderKernelApp(IPKernelApp):
+
+        if IPYKERNEL_6:
+            outstream_class = DottedObjectName(
+                'spyder_kernels.console.outstream.TTYOutStream')
+
+        def init_pdb(self):
+            """
+            This method was added in IPykernel 5.3.1 and it replaces
+            the debugger used by the kernel with a new class
+            introduced in IPython 7.15 during kernel's initialization.
+            Therefore, it doesn't allow us to use our debugger.
+            """
+            pass
+
+    # Fire up the kernel instance.
+    kernel = SpyderKernelApp.instance()
     kernel.kernel_class = SpyderKernel
     try:
         kernel.config = kernel_config()
